// -*- mode: js; js-indent-level: 4; indent-tabs-mode: nil -*-
/* Copyright 2012 Sam Bull */

const GLib = imports.gi.GLib;
const Gtk = imports.gi.Gtk;
const Lang = imports.lang;
const Mainloop = imports.mainloop;
const Signals = imports.signals;
const St = imports.gi.St;

const IconGrid = imports.ui.iconGrid;
const Main = imports.ui.main;
const Overview = imports.ui.overview;
const Tweener = imports.ui.tweener;

const PATH = GLib.get_home_dir() + "/.local/share/gnome-shell/.notes"

/* This is the page object to create our notes view. */
const NotesView = new Lang.Class({
    Name: 'NotesView',

    _init: function() {
        // Use IconGrid from AppsDisplay to arrange the notes.
        this._grid = new IconGrid.IconGrid({ xAlign: St.Align.START });
        this._grid.actor.connect('style-changed',
                                 Lang.bind(this._grid, _onStyleChanged));

        this.actor = new St.ScrollView({ x_fill: true,
                                         y_fill: false,
                                         y_align: St.Align.START,
                                         style_class: 'vfade' });
        this.actor.add_actor(this._grid.actor);
        this.actor.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
        this.actor.connect('notify::mapped', Lang.bind(this,
            function() {
                if (!this.actor.mapped)
                    return;

                let adjustment = this.actor.vscroll.adjustment;
                let direction = Overview.SwipeScrollDirection.VERTICAL;
                Main.overview.setScrollAdjustment(adjustment, direction);

                // Reset scroll on mapping
                adjustment.value = 0;
            }));

        this._workId = Main.initializeDeferredWork(
            this.actor, Lang.bind(this, this._redisplay));
    },

    _redisplay: function() {
        // Read contents from file
        let content;
        try {
            content = GLib.file_get_contents(PATH)[1];
        } catch (e) {
            content = "";
        }

        // Split individual notes using a private unicode character.
        content = content.toString().split("\uE000");

        this._grid.removeAll();
        this.notes = new Array();
        // Initialise loaded notes with a trailing blank note.
        let note;
        for (let i=0; i < content.length; i++) {
            note = this.addNote(content[i]);
        }
        if (content[content.length-1].length > 0) {
            note = this.addNote("");
        }
        // Attach signal to last, empty note.
        this.signal_note = note._entry.get_clutter_text().connect(
            'text-changed', Lang.bind(this, this.addNote));
    },

    // Callback for creating new note, so there's always an empty one.
    // init is used to initialise the strings on initial load.
    addNote: function(init) {
        let new_note = false;
        if (typeof init !== 'string') {
            let ct = this.notes[this.notes.length-1]._entry.get_clutter_text()
            ct.disconnect(this.signal_note);
            init = "";
            new_note = true;
        }

        // Connect a close button to the previous note.
        if (this.notes.length) {
            let prev_note = this.notes[this.notes.length-1]
            prev_note.actor.connect('enter-event', Lang.bind(
                prev_note, prev_note._onEnter));
            prev_note.actor.connect('leave-event', Lang.bind(
                prev_note, prev_note._onLeave));
            prev_note.connect('show-close-button',
                              Lang.bind(this, this._onButtonClose));
        }

        let note = new NoteIcon(this, init);
        this.notes.push(note);
        this._grid.addItem(note.actor);

        if (new_note) {
            this.signal_note = note._entry.get_clutter_text().connect(
                'text-changed', Lang.bind(this, this.addNote));

            note.actor.set_scale(0,0);
            Tweener.addTween(note.actor, {time: 1, transition: 'easeOutQuad',
                                          scale_x: 1, scale_y: 1});
        }

        return note;
    },

    // Remove note after close button has been clicked.
    delNote: function(button) {
        for (let i=0; i < this.notes.length; i++) {
            if (this.notes[i] == button._note) {
                let note = this.notes[i];
                Tweener.addTween(note.actor, {time: 1,
                                              transition: 'easeOutQuad',
                                              scale_x: 0, scale_y: 0});
                this.notes.splice(i, 1);
                Mainloop.timeout_add(1000, Lang.bind(this, function () {
                        this.save();
                        this._redisplay();
                    }));
                break;
            }
        }
    },

    // Save notes to file.
    save: function() {
        let content = "";
        for (let i=0; i < this.notes.length; i++) {
            let note = this.notes[i];
            let text = "";
            if (!note._title.visible)
                text = note._entry.get_text();
            else if (note._title.get_text().length > 0 ||
                note._entry.get_text().length > 0)
                text = note._title.get_text() + "\n" + note._entry.get_text();

            content = content.concat("\uE000", text);
        }

        let byteData = imports.byteArray.fromString(content.slice(1), "UTF-8");
        GLib.file_set_contents(PATH, byteData);
    },

    show: function() {
        for (let i=0; i < this.notes.length; i++)
            this.notes[i].show();
    },

    hide: function() {
        for (let i=0; i < this.notes.length; i++)
            this.notes[i].hide();
    },

    // Hide other buttons when a button is shown.
    _onButtonClose: function (note) {
        for (let i=0; i < this.notes.length; i++) {
            let n = this.notes[i];
            if (n == note)
                continue;
            n.hideCloseButton();
        }
    }
});

/* Increase size of IconGrid items to allow full sized notes. */
function _onStyleChanged() {
    let themeNode = this.actor.get_theme_node();
    this._spacing = themeNode.get_length('spacing');
    this._hItemSize = 200;
    this._vItemSize = 280;
    this._grid.queue_relayout();
}

/* The note item to be added to IconGrid. */
const NoteIcon = new Lang.Class({
    Name: 'NoteIcon',

    _init : function(notesView, text) {
        this.notesView = notesView;
        let i = text.indexOf("\n");
        if (i != -1)
            text = [text.slice(0,i), text.slice(i+1)];
        else
            text = [text, ""];

        this._entry = new St.Entry();
        this._entry.set_text(text[1]);
        this._title = new St.Label({ style_class: 'title' });
        this._title.set_text(text[0]);

        this.actor = new St.BoxLayout({ style_class: 'note' });
        this.actor.set_reactive(true);
        this.actor.set_vertical(true);
        this.actor._delegate = this;
        this.actor.add(this._title, {y_fill: false, y_align: St.Align.START});
        this.actor.add(this._entry, {y_fill: false, y_align: St.Align.START});

        // Tweak text entry properties to better suit notes.
        let clutterText = this._entry.get_clutter_text();
        clutterText.set_single_line_mode(false);
        clutterText.set_activatable(false);
        clutterText.set_line_wrap(true);
        clutterText.set_line_wrap_mode(imports.gi.Pango.WrapMode.WORD_CHAR);

        // Close button
        this._idleToggleCloseId = 0;
        this.button = new St.Button({ style_class: 'window-close' });
        this.button._overlap = 0;
        this.button._note = this;
        this.button.hide();
        this.button.set_position(180,-10);
        this.button.connect('clicked', Lang.bind(notesView, notesView.delNote));
        this.actor.add(this.button, {y_fill: false, y_align: St.Align.START});
        this.button.connect('style-changed',
                            Lang.bind(this, this._onStyleChanged));

        // Make sure key entry is accepted after clicking anywhere on actor.
        this.actor.connect('button-press-event', function () {
            clutterText.grab_key_focus();
        });

        // Save data when focus leaves widget.
        clutterText.connect('key-focus-out', Lang.bind(this.notesView,
                                                       this.notesView.save));
        // Handle displaying and editing of title.
        clutterText.connect('key-focus-in', Lang.bind(this, this._title_in));
        clutterText.connect('key-focus-out', Lang.bind(this, this._title_out));
    },


    // Hide label, and stick title into entry so it's editable.
    _title_in: function() {
        this._title.hide();
        if (this._title.get_text().length > 0 ||
            this._entry.get_text().length > 0) {
            this._entry.set_text(this._title.get_text() + "\n" +
                                 this._entry.get_text());
        }
    },

    // Strip out title, and reshow the title label.
    _title_out: function() {
        let text = this._entry.get_text();
        if (text.length > 0) {
            let i = text.indexOf("\n");
            if (i != -1)
                text = [text.slice(0,i), text.slice(i+1)];
            else
                text = [text, ""];
            this._title.set_text(text[0]);
            this._entry.set_text(text[1]);
        } else
            this._title.set_text("");
        this._title.show();
    },

    hide: function() {
        this.button.hide();
    },

    show: function() {
        if (this._entry.has_pointer)
            this.button.show();
    },

    _onStyleChanged: function() {
        let closeNode = this.button.get_theme_node();
        this.button._overlap = closeNode.get_length('-shell-close-overlap');

        this.notesView._grid.actor.queue_relayout();
    },

    _onEnter: function() {
        this.button.show();
        this.button.raise_top();
        this.emit('show-close-button');
    },

    _onLeave: function() {
        if (this._idleToggleCloseId == 0)
            this._idleToggleCloseId = Mainloop.timeout_add(
                750, Lang.bind(this, this._idleToggleCloseButton));
    },

    _idleToggleCloseButton: function() {
        this._idleToggleCloseId = 0;
        if (!this._entry.get_clutter_text().has_pointer &&
            !this.button.has_pointer && !this.actor.has_pointer)
            this.button.hide();

        return false;
    },

    hideCloseButton: function() {
        if (this._idleToggleCloseId > 0) {
            Mainloop.source_remove(this._idleToggleCloseId);
            this._idleToggleCloseId = 0;
        }
        this.button.hide();
    }
});
Signals.addSignalMethods(NoteIcon.prototype);
